/* 
 *   Creation Date: <2001/10/14 04:05:37 samuel>
 *   Time-stamp: <2001/10/20 14:25:29 samuel>
 *   
 *	<main.c>
 *	
 *	User process/kernel module interface 
 *   
 *   Copyright (C) 2001 Samuel Rydh (samuel@ibrium.se)
 *   
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation
 *   
 */

#include "mol_config.h"

#include <sys/mman.h>
#include <sys/time.h>
#include "molcpu.h"
#include "version.h"
#include "memory.h"

/* Assembly export (from mainloop_asm.S) */
extern void register_asm_rvecs( void );


/************************************************************************/
/*	Initialization / Cleanup					*/
/************************************************************************/

void 
molcpu_arch_init( void )
{
	register_asm_rvecs();
}

void 
molcpu_arch_cleanup( void )
{
}

void
map_mregs( void )
{
	mregs = (mac_regs_t*) map_phys_mem( 0, _get_mregs_phys(), sizeof(mac_regs_t), 
					    PROT_READ | PROT_WRITE );
	if( !mregs ) {
		perrorm("map_kmem:");
		exit(1);
	}
	memset( mregs, 0, sizeof(mac_regs_t) );
}

void
unmap_mregs( void )
{
	if( mregs )
		unmap_mem( (char*)mregs, sizeof(mac_regs_t) );
}


/************************************************************************/
/*	Create MOL session						*/
/************************************************************************/

void
check_kmod_version( void )
{
	int kvers;

	/* Make sure the kernel module is up to date... */
	kvers = _get_mol_mod_version();
	if( kvers < 0 ){
		fprintf( stderr, "Error: The MOL kernel module is not loaded\n" );
		exit(1);
	}
	if( kvers != MOL_VERSION ){
		printm("The MOL kernel module version %d.%d.%d does not match the binary.\n",
		       (kvers>>16), (kvers>>8)&0xff, kvers&0xff );
		printm("Please unload the kernel module (rmmod mol) and try again\n");
		exit(1);
	}
	if( is_a_601() ){
		printm("Mac-on-Linux no longer supports the PowerPC 601 processor\n");
		exit(1);
	}
}

int 
open_session( void )
{
	struct timeval tv;
	int err;

	gettimeofday( &tv, NULL );
	srandom( tv.tv_usec );
	g_session_magic = random();

	if( (err=_kernel_init( g_session_magic )) ) {
		if( err == errSessionInUse ) {
			printm("---> Removing unreleased MOL session\n");
			_kernel_cleanup();
			err = _kernel_init( g_session_magic );
		}
		if( err == errSessionOutOfBounds )
			printm("The session number (%d) is out of bounds\n", g_session_id );
		else if( err )
			printm("_kernel_init returned an error %d\n", err);
	}
	return err;
}
